// Copyright (c) 2024 Alfio Greco 

//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with ART.  If not, see <http://www.gnu.org/licenses/>.



// @ART-colorspace: "rec2020"
// @ART-label: "sigmoid ycbcr"


float fmin(float a, float b)
{
    if (a < b) {
        return a;
    } else {
        return b;
    }
}



float[3] vibrance_s(varying float r, varying float g,varying float b,
        float vibrance)
{
    float rgb[3] = {r, g, b};
    

    float lum = rgb[0] * 0.292 + rgb[1] * 0.797 + rgb[2] * 0.046;

    if (lum <= 0) { 
        lum = 0.00001;
        }
    else { 
        lum=lum;
        }	


    float minvalue;
    minvalue= fmin(fmin (rgb[0],rgb[1]),rgb[2]);




	
    float saturation = 1 - (minvalue/lum);
 

    float mask;	
    float original[3];		
    float saturation2;	
    float mult;
	
    if (saturation == 0) { 
        saturation = 0.00001;
        }
    else { 
        saturation=saturation;
        }	


    if (saturation > 1) { 
        mask = 1;
        }
    else { 
        mask=saturation;
        }  
		
    if (mask < 0) { 
        mask = 0;
        }
    else { 
        mask=mask;
        }  



		
    original[0]=rgb[0];
	original[1]=rgb[1];
	original[2]=rgb[2];	
	
	
	rgb[0]=(rgb[0]-lum)*vibrance+lum;
	rgb[1]=(rgb[1]-lum)*vibrance+lum;
	rgb[2]=(rgb[2]-lum)*vibrance+lum;	


    if (vibrance > 1) { 
        	rgb[0]=rgb[0]*(1-mask)+original[0]*mask;
	        rgb[1]=rgb[1]*(1-mask)+original[1]*mask;
	        rgb[2]=rgb[2]*(1-mask)+original[2]*mask;
        }
    else { 
       	    rgb[0]=rgb[0]*(mask)+original[0]*(1-mask);
	        rgb[1]=rgb[1]*(mask)+original[1]*(1-mask);
	        rgb[2]=rgb[2]*(mask)+original[2]*(1-mask);
        }  	


    minvalue= fmin(fmin (rgb[0],rgb[1]),rgb[2]);




	
    saturation = 1 - (minvalue/lum);
	
	
    float param_choice=0.89;

	
    if (saturation <= param_choice) { 
        saturation2 = saturation;
        }
    else { 
        saturation2=param_choice;
        }    

    mult=saturation2 / saturation;
	
	
	rgb[0]=(rgb[0]-lum)*mult+lum;
	rgb[1]=(rgb[1]-lum)*mult+lum;
	rgb[2]=(rgb[2]-lum)*mult+lum;	
	
	
	
	
    return rgb;
}

float[3] rec2020_to_oklab(varying float r, varying float g,varying float b)
{
    float rgb[3] = {r, g, b};
    float rtemp=rgb[0];
    float gtemp=rgb[1];
    float btemp=rgb[2];	
    const float gamma = 3;
    const float igamma = 1 / gamma;	
	
//rec2020 to xyz
	rgb[0]=rtemp * 0.6369580 + gtemp * 0.1446169 + btemp * 0.1688810;
	rgb[1]=rtemp * 0.2627002 + gtemp * 0.6779981 + btemp * 0.0593017;
	rgb[2]=gtemp * 0.0280727 + btemp * 1.0609851;
	
//xyz to lms	

	rtemp=rgb[0] * 0.8189330101 + rgb[1] * 0.3618667424 - rgb[2] * 0.1288597137;
	gtemp=rgb[0] * 0.0329845436 + rgb[1] * 0.9293118715 + rgb[2] * 0.0361456387;
	btemp=rgb[0] * 0.0482003018 + rgb[1] * 0.2643662691 + rgb[2] * 0.6338517070;

	rtemp=pow(rtemp,igamma);
	gtemp=pow(gtemp,igamma);
	btemp=pow(btemp,igamma);


//lms to oklab

	rgb[0]=rtemp * 0.2104542553 + gtemp * 0.7936177850 - btemp * 0.0040720468;
	rgb[1]=rtemp * 1.9779984951 - gtemp * 2.4285922050 + btemp * 0.4505937099;
	rgb[2]=rtemp * 0.0259040371 + gtemp * 0.7827717662 - btemp * 0.8086757660;

    return rgb;

}

float[3] oklab_to_rec2020(varying float l, varying float a,varying float b)
{
    float lab[3] = {l, a, b};
    float ltemp=lab[0];
    float atemp=lab[1];
    float btemp=lab[2];	
    const float gamma = 3;
    const float igamma = 1 / gamma;	
	
	
	
//oklab to lms
	lab[0]=ltemp * 1 + atemp * 0.39633 + btemp * 0.21580;
	lab[1]=ltemp * 1 - atemp * 0.10556 - btemp * 0.06385;
	lab[2]=ltemp * 1 - atemp * 0.08948 - btemp * 1.29148;

	lab[0]=pow(lab[0],gamma);
	lab[1]=pow(lab[1],gamma);
	lab[2]=pow(lab[2],gamma);
	
//lms to xyz	

	ltemp=lab[0] * 1.22701 - lab[1] * 0.55779 + lab[2] * 0.28125;
	atemp=lab[0] * -0.04058 + lab[1] * 1.11225 - lab[2] * 0.07167;
	btemp=lab[0] * -0.07638 - lab[1] * 0.42148 + lab[2] * 1.58616;

//xyz to rec2020

	lab[0]=ltemp * 1.7166512 - atemp * 0.3556708 - btemp * 0.2533663;
	lab[1]=ltemp * -0.6666844 + atemp * 1.6164812 + btemp * 0.0157685;
	lab[2]=ltemp * 0.0176399 - atemp * 0.0427706 + btemp * 0.9421031;

    return lab;

}

float[3] gamut_limiter(varying float r, varying float g,varying float b,
        float param_choice)
{
    float rgb[3] = {r, g, b};
    

    float lum = rgb[0] * 0.25 + rgb[1] * 0.50 + rgb[2] * 0.25;

    if (lum <= 0) { 
        lum = 0.00001;
        }
    else { 
        lum=lum;
        }	


    float minvalue;
    minvalue= fmin(fmin (rgb[0],rgb[1]),rgb[2]);




	
    float saturation = 1 - (minvalue/lum);
    float saturation2;	
    float mult;
	
    if (saturation == 0) { 
        saturation = 0.00001;
        }
    else { 
        saturation=saturation;
        }	

	
    if (saturation <= 0.99999) { 
        saturation2 = saturation;
        }
    else { 
        saturation2=0.99999;
        }    

    mult=saturation2 / saturation;
	
	
	rgb[0]=(rgb[0]-lum)*mult+lum;
	rgb[1]=(rgb[1]-lum)*mult+lum;
	rgb[2]=(rgb[2]-lum)*mult+lum;		


    float oklab[3];
    float chroma;
	
	oklab=rec2020_to_oklab(rgb[0],rgb[1],rgb[2]);
	chroma= sqrt ((oklab[1]*oklab[1])+(oklab[2]*oklab[2]));
	
	


	
    if (saturation <= param_choice) { 
        saturation2 = saturation;
        }
    else { 
        saturation2=param_choice;
        }    

    mult=saturation2 / saturation;
	
	
	rgb[0]=(rgb[0]-lum)*mult+lum;
	rgb[1]=(rgb[1]-lum)*mult+lum;
	rgb[2]=(rgb[2]-lum)*mult+lum;	
	
	
    float oklab2[3];
    float chroma2;
    float chroma_mult;
    float rgb2[3];	

	
	oklab2=rec2020_to_oklab(rgb[0],rgb[1],rgb[2]);	
	chroma2= sqrt ((oklab2[1]*oklab2[1])+(oklab2[2]*oklab2[2]));
	
	chroma_mult= chroma2/chroma;
	
	oklab2[0]=oklab2[0];
	oklab2[1]=oklab[1]*chroma_mult;
	oklab2[2]=oklab[2]*chroma_mult;	


	
	rgb2=oklab_to_rec2020(oklab2[0],oklab2[1],oklab2[2]);
	
	rgb[0]=rgb[0]*0.75+rgb2[0]*0.25;
	rgb[1]=rgb[1]*0.75+rgb2[1]*0.25;
	rgb[2]=rgb[2]*0.75+rgb2[2]*0.25;	
	
	
	
    return rgb;
}


float[3] sigmoid( float r,  float g, float b,
        float contrast, float power, float vibrance, float luma_r, float hue)
{
    float rgb[3] = {r, g, b};
    const float gamma = 3;
    const float igamma = 1 / gamma;	
    
    const float gray = 0.18;   
    const float pivot = (1-gray)/gray; 
	
	float luma;	
	float kr=0.292;	
	float kg=0.797;		
	float kb=0.046;		
	luma=rgb[0]*kr+rgb[1]*kg+rgb[2]*kb;	
	float cb0;
	float cr0;

	cb0=(rgb[2]-luma)/(1-kb);	
	cr0=(rgb[0]-luma)/(1-kr);
	float croma0;
    croma0= sqrt(cb0*cb0+cr0*cr0);
	
	rgb[0]=pow(rgb[0],power);
	rgb[1]=pow(rgb[1],power);
	rgb[2]=pow(rgb[2],power);

	rgb=vibrance_s(rgb[0],rgb[1],rgb[2],vibrance);



	
	
	luma=pow(luma,power);	
    luma = (1-(pivot/(pow((luma/gray),contrast)+pivot)));


	
	rgb[0]=(1-(pivot/(pow((rgb[0]/gray),contrast)+pivot)));
	rgb[1]=(1-(pivot/(pow((rgb[1]/gray),contrast)+pivot)));
	rgb[2]=(1-(pivot/(pow((rgb[2]/gray),contrast)+pivot)));


	float lumay;
	float cb;
	float cr;
	
	lumay=rgb[0]*kr+rgb[1]*kg+rgb[2]*kb;
	cb=(rgb[2]-lumay)/(1-kb);	
	cr=(rgb[0]-lumay)/(1-kr);
	
	float croma;
	float cmult;	
    croma= sqrt(cb*cb+cr*cr);	
	cmult=croma/croma0;
	
	float cbh;
	float crh;	

    cbh=cb0*cmult;
    crh=cr0*cmult;
	

    lumay=luma*luma_r+lumay*(1-luma_r);



    float rgbh[3];	
	rgbh[0]=crh*(1-kr)+lumay;
	rgbh[2]=cbh*(1-kb)+lumay;
	rgbh[1]=(lumay-(rgbh[2]*kb)-(rgbh[0]*kr))/kg;		
	
	
	
	rgb[0]=cr*(1-kr)+lumay;
	rgb[2]=cb*(1-kb)+lumay;
	rgb[1]=(lumay-(rgb[2]*kb)-(rgb[0]*kr))/kg;	

	rgb[0]=rgbh[0]*hue+rgb[0]*(1-hue);
	rgb[1]=rgbh[1]*hue+rgb[1]*(1-hue);
	rgb[2]=rgbh[2]*hue+rgb[2]*(1-hue);

    return rgb;
	
}
               

// @ART-param: ["param_choice", "rec2020 gamut limit", 0.0, 1.0, 0.89, 0.01]
// @ART-param: ["expcomp", "Exposure compensation", -5.0, 5.0, 0.95, 0.01]
// @ART-param: ["power", "brightness", 0.0, 4.0, 1.00, 0.01]
// @ART-param: ["contrast", "contrast", 0.0, 4.0, 1.40, 0.01]
// @ART-param: ["vibrance", "vibrance", 0.0, 3.0, 1.00, 0.01]
// @ART-param: ["hue", "restore ycbcr hue", 0.0, 1.0, 0.13, 0.01]
// @ART-param: ["luma_r", "restore ycbcr luma", 0.0, 1.0, 0.50, 0.01]


void ART_main(varying float r, varying float g, varying float b,             
              output varying float rout,
              output varying float gout,
              output varying float bout,
              float param_choice,float expcomp,
			  float contrast,
			  float power,
			  float vibrance,
			  float luma_r,
			  float hue)
{
    float rgb[3];
	
    const float scale = pow(2, expcomp);



    rgb = gamut_limiter(r, g, b, param_choice);

	rgb[0]=rgb[0] * scale;
	rgb[1]=rgb[1] * scale;
	rgb[2]=rgb[2] * scale;	

    rgb = sigmoid(rgb[0], rgb[01], rgb[2], contrast, power,vibrance, luma_r, hue);

    rout = rgb[0];
    gout = rgb[1];
    bout = rgb[2];
}
